use crate::vm::types::DmType;
use crate::vm::DmValue::*;
use crate::vm::{DmObject, DmValue, DmVirtualMachine};
use DmType::*;
use crate::vm::module::DmFunction;

fn get_method<'a>(ptr: *mut DmObject, name: String) -> Option<&'a DmFunction> {
    unsafe {
        (*ptr).object_type.methods.get(&name)
    }
}

fn get_rust_string_from_dm_object(ptr: *mut DmObject) -> String {
    unsafe {
        let dm_object = Box::from_raw(ptr);
        format!("{}@{:?}", dm_object.object_type.name, dm_object.data)
    }
}

fn get_rust_string_from_dm_string(ptr: *mut DmObject) -> String {
    unsafe {
        let dm_string = Box::from_raw(ptr);
        todo!()
    }
}

pub fn dm_print(args: Vec<DmValue>, vm: &mut DmVirtualMachine) -> DmValue {
    if args.len() != 1 {
        return DmException(todo!("make an Exception object."))
    }
    match args[0] {
        DmInt(i) => {
            print!("{}", i);
        }
        DmLong(l) => {
            print!("{}", l);
        }
        DmDouble(d) => {
            print!("{}", d);
        }
        DmPointer(ptr) => {
            if let Some(to_string) = get_method(ptr, String::from("to_string")) {
                let call_result = vm.call(&to_string, vec![args[0]]);
                match call_result {
                    DmPointer(dm_string_ptr) => {
                        let string = get_rust_string_from_dm_string(dm_string_ptr);
                        print!("{}", string);
                    }
                    _ => {
                        // TODO: vm throw or return exception?
                    }
                }
            } else {
                print!("{}", get_rust_string_from_dm_object(ptr));
            }
        },
        DmUnit => {
            print!("Unit")
        },
        DmException(e) => {
            print!("Exception")
        }
    }
    DmUnit
}
